from . import components as components
from .components import Cluster
from .components import Cluster_Node
from .components import Session

import math 

class Clustering_Engine():
    def __init__(self):
        return

    ###DESCRIPTION: A 2-layer topology is a tree-like topology in which the root node is the roo_aggregator, 
    # and level_1 leaves are aggregators or aggregator_trainers, and the level_2 leaves are trainer-only nodes. 
    # Example::
    ##                    [AGG____0]
    ##                   /     |     \
    ##              [AGG_1]   ...     [AGG_n]
    ##             /   |   \         /   |   \
    ##           [T1] [T2] [T3]  [Tk-2] [Tk-1] [Tk]
    ###_____________________________________________________________________________________________________
    ###_____________________________________________________________________________________________________
    
    def create_central_aggregation_topology(self,session): #TODO:incorporate 30,70 or 20,80 or ...
        if(len(session.client_list) < 4):
            print("Number of clients is not enough for a 2-layer hierarchical clustering.")
            return
        
        session.role_vector = []
        session.role_dictionary = {}
        num_aggregators = 1

        num_training_only = session.session_capacity_max - num_aggregators
        
        session.role_dictionary['agg_0'+ "_" + str(session.session_id)] = []
        session.role_vector.append(0)
        n_counter = 0
        for j in range(num_training_only):
            if(n_counter >= num_training_only):
                break
            else:
                session.role_dictionary['agg_0'+ "_" + str(session.session_id)].append('t_'+str(n_counter)+ "_" + str(session.session_id))     
                # session.role_dictionary['agg_' + str(i) + "_" + str(session.session_id)].append('t_'+str(n_counter)+ "_" + str(session.session_id))
                n_counter += 1
            

        return [session.role_vector,session.role_dictionary]

    def create_2layer_topology(self,session,percentage_of_aggs): #TODO:incorporate 30,70 or 20,80 or ...
        if(len(session.client_list) < 4):
            print("Number of clients is not enough for a 2-layer hierarchical clustering.")
            return
        
        session.role_vector = []
        session.role_dictionary = {}
        num_aggregators = math.floor(session.session_capacity_max * percentage_of_aggs)
        if(num_aggregators < 2):
            num_aggregators = 2

        num_training_only = session.session_capacity_max - num_aggregators
        num_trainer_per_l2_cluster = math.floor(num_training_only / (num_aggregators - 1))
        
        session.role_dictionary['agg_0'+ "_" + str(session.session_id)] = []
        session.role_vector.append(0)
        n_counter = 0
        for i in range(1,num_aggregators):
            session.role_dictionary['agg_0'+ "_" + str(session.session_id)].append('agg_' + str(i)+ "_" + str(session.session_id))
            session.role_vector.append(0)  
            session.role_dictionary['agg_' + str(i)+ "_" + str(session.session_id)] = []
            for j in range(num_trainer_per_l2_cluster):
                if(n_counter >= num_training_only):
                    break
                else:
                    session.role_dictionary['agg_' + str(i) + "_" + str(session.session_id)].append('t_'+str(n_counter)+ "_" + str(session.session_id))
                    n_counter += 1
            
        if(n_counter < num_training_only):#Distribute the remaining nodes to the aggregators, each receiving one of the remaining nodes
            for z in range(1,num_aggregators):
                if(n_counter >= num_training_only):
                    break
                else:
                    session.role_dictionary['agg_' + str(z) + "_" + str(session.session_id)].append('t_'+str(n_counter)+ "_" + str(session.session_id))
                    n_counter += 1

        return [session.role_vector,session.role_dictionary]
 
    def greedy_based_topology(self,session):
        #TODO: Read the calculated cost of FL in the previous round, and accordingly build a new role_vector, and feed it to the update_roles.
        session.update_roles([])
        return []

    def form_clusters(self,session):
        items = list(session.role_dictionary.keys())#check session.role_dictionary
        for i in range(len(items)):
            new_cluster = Cluster('cluster_' + str(i))

            new_node = None
            has_node = False
            for n in session.nodes:
                if(n.role == str(items[i])):
                    # print("already has node")
                    new_node = n
                    has_node = True
                    break
            if(has_node == False):
                new_node = Cluster_Node("N_" + str(len(session.nodes)),items[i],None) #First create the cluster head which has the role of aggregator
                new_node.status = components._NODE_PENDING
                session.nodes.append(new_node)#In each session, there is one root node which is the top-most aggregator. in role_dic it is 'agg_0'. If a given node is 'agg_0', then it is root node
            
            if(items[i] == "agg_0_" + str(session.session_id)):
                # new_node.role = components._ROLE_AGGREGATOR_ROOT
                session.set_root_node(new_node)
            
            new_cluster.set_cluster_head(new_node)
            
            for j in range(0,len(session.role_dictionary[items[i]])):#Now form clusters of nodes (not clients) based on the list of each aggregator's items
                new_sub_node = Cluster_Node("N_" + str(len(session.nodes)),session.role_dictionary[items[i]][j],None)
                new_sub_node.status = components._NODE_PENDING
                session.nodes.append(new_sub_node)
                # if(items[i][j][0] == 't'):
                #     new_sub_node.role = components._ROLE_TRAINER
                new_cluster.add_node(new_sub_node)
                
            session.add_cluster(new_cluster)
        #NOTE: All added nodes here are in pending mode. They will go to ACTIVE mode once the client they are assigned to, acknowledges the node's role.

        
        
        


import torch
import torch.nn as nn
import torch.optim as optim
import torchvision
import torchvision.transforms as transforms
import random
from Core.sdflmq_client_logic import SDFLMQ_Client
import random
import time
from datetime import datetime, timedelta
from torch.utils.data import Subset
# Define device
device = torch.device("cuda" if torch.cuda.is_available() else "cpu")

# Load CIFAR-10 dataset
transform = transforms.Compose([
    transforms.ToTensor(),
    transforms.Normalize((0.5, 0.5, 0.5), (0.5, 0.5, 0.5))
])

trainset = torchvision.datasets.CIFAR10(root='./data', train=True, download=True, transform=transform)

# Shuffle dataset before sampling
indices = list(range(len(trainset)))
random.shuffle(indices)
subset_size = int(0.1 * len(trainset))  # 10% of the dataset
subset_indices = indices[:subset_size]
train_subset = torch.utils.data.Subset(trainset, subset_indices)

trainloader = torch.utils.data.DataLoader(train_subset, batch_size=64, shuffle=True, num_workers=2)

testset = torchvision.datasets.CIFAR10(root='./data', train=False, download=True, transform=transform)
testloader = torch.utils.data.DataLoader(testset, batch_size=64, shuffle=False, num_workers=2)

# Define the CNN model
class CNN(nn.Module):
    def __init__(self):
        super(CNN, self).__init__()
        self.conv1 = nn.Conv2d(3, 64, kernel_size=3, padding=1)
        self.conv2 = nn.Conv2d(64, 64, kernel_size=3, padding=1)
        self.pool = nn.MaxPool2d(kernel_size=2, stride=2)
        self.conv3 = nn.Conv2d(64, 128, kernel_size=3, padding=1)
        self.conv4 = nn.Conv2d(128, 128, kernel_size=3, padding=1)
        self.conv5 = nn.Conv2d(128, 256, kernel_size=3, padding=1)
        self.conv6 = nn.Conv2d(256, 256, kernel_size=3, padding=1)
        self.fc1 = nn.Linear(4 * 4 * 256, 512)
        self.fc2 = nn.Linear(512, 256)
        self.fc3 = nn.Linear(256, 10)
        self.dropout = nn.Dropout(0.5)

    def forward(self, x):
        x = torch.relu(self.conv1(x))
        x = torch.relu(self.conv2(x))
        x = self.pool(x)
        x = torch.relu(self.conv3(x))
        x = torch.relu(self.conv4(x))
        x = self.pool(x)
        x = torch.relu(self.conv5(x))
        x = torch.relu(self.conv6(x))
        x = self.pool(x)
        x = x.view(-1, 4 * 4 * 256)
        x = torch.relu(self.fc1(x))
        x = self.dropout(x)
        x = torch.relu(self.fc2(x))
        x = self.dropout(x)
        x = self.fc3(x)
        return x
    
FL_ROUNDS = 50
session_name = "session_14_cifar"
# Create model, loss function, and optimizer
model = CNN().to(device)
criterion = nn.CrossEntropyLoss()
optimizer = optim.Adam(model.parameters(), lr=0.001)
def printsomething():
        print("Model update received")
myid = "client_" + str(random.randint(0,10000)) 
fl_client = SDFLMQ_Client(  myID=myid,
                                broker_ip = 'localhost',
                                broker_port = 1883,
                                preferred_role="aggregator",
                                loop_forever=False)
fl_client.join_fl_session(session_id=session_name,
                                fl_rounds=FL_ROUNDS,
                                model_name="cifar10_cnn",
                                preferred_role="aggregator")



for k in range(FL_ROUNDS):
    print("Round " + str(k))
    # # Training Loop
    # num_epochs = 2
    # for epoch in range(num_epochs):
    #     running_loss = 0.0
    #     for inputs, labels in trainloader:
    #         inputs, labels = inputs.to(device), labels.to(device)
    #         optimizer.zero_grad()
    #         outputs = model(inputs)
    #         loss = criterion(outputs, labels)
    #         loss.backward()
    #         optimizer.step()
    #         running_loss += loss.item()
    #     print(f"Epoch {epoch+1}, Loss: {running_loss/len(trainloader):.4f}")
    
    #Sending the Model for Aggregation    
    fl_client.set_model(session_name,model)
    fl_client.send_local(session_name)
    fl_client.wait_global_update()


    # # Test phase using 100 samples from the test dataset
    # test_subset_size = 100
    # test_indices = list(range(len(testset)))
    # random.shuffle(test_indices)
    # test_subset_indices = test_indices[:test_subset_size]
    # test_subset = torch.utils.data.Subset(testset, test_subset_indices)
    # testloader_subset = torch.utils.data.DataLoader(test_subset, batch_size=100, shuffle=False)

    # correct = 0
    # total = 0
    # model.eval()
    # with torch.no_grad():
    #     for inputs, labels in testloader_subset:
    #         inputs, labels = inputs.to(device), labels.to(device)
    #         outputs = model(inputs)
    #         _, predicted = torch.max(outputs, 1)
    #         total += labels.size(0)
    #         correct += (predicted == labels).sum().item()

    # print(f"Test Accuracy on 100 samples: {100 * correct / total:.2f}%")



import os
class Client:
    name = ""
    id = ""
    directory = ""
    broker = ""
    status = ""
    executables = [""]

    def __init__(self, name, id, type, directory, broker,executables=[]):
        self.name = name
        self.id = id
        self.type = type
        self.directory = directory
        self. broker = broker
        self.status = "OFF"
        self.executables = executables

    def power_on(self):
        return 0
    
    def power_off(self):
        return 0

    def check_status(self):
        return 0
    
    def update_status(self, new_status):
        self.status = new_status
    




import numpy as np
import datetime
import matplotlib.pyplot as plt
from . import components as components 
import os
import json
from ..Coordinator_Modules.components import Cluster
from ..Coordinator_Modules.components import Cluster_Node
from ..Coordinator_Modules.components import Session
from ..Coordinator_Modules.components import Client

class Session_Manager():
    def __init__(self):
        self.__sessions = {}
    
    def All_Nodes_Ready(self,session_id):
        active_nodes = 0
        for n in self.__sessions[session_id].nodes:
            # if(n.is_elected):
                if(n.status == components._NODE_ACTIVE):
                    active_nodes += 1
        # if(active_nodes == len(self.__sessions[session_id].nodes)):
        if((active_nodes == self.__sessions[session_id].session_capacity_max) or 
           ((active_nodes >= self.__sessions[session_id].session_capacity_min) and 
            (datetime.datetime.now() > self.__sessions[session_id].get_current_round()['starting_time'] + self.__sessions[session_id].waiting_time))):
            return [True, active_nodes]
        else:
            return [False, active_nodes]
   
    def update_session(self,session_id):

        # print("Current round index: " + str(self.__sessions[session_id].current_round_index))
        # print("Max rounds: " + str(self.__sessions[session_id].num_rounds))
        
        if(self.__sessions[session_id].current_round_index >= self.__sessions[session_id].num_rounds):
            # print("Session terminated for reaching max fl rounds")
            self.__sessions[session_id].session_status = components._SESSION_TERMINATED
            return

        if(self.__sessions[session_id].session_creation_time + #Check session Time
           self.__sessions[session_id].session_time < datetime.datetime.now()):
            if((len(self.__sessions[session_id].client_list) >= self.__sessions[session_id].session_capacity_min)): #Check list of clients, in relation to min capacity and max capacity
                self.__sessions[session_id].session_status = components._SESSION_ACTIVE  #if greater than min cap is met then session ready
                return
                # print("Session ready")
            else:
                self.__sessions[session_id].session_status = components._SESSION_TIMEOUT
                return
                # print("Session with session_id " + session_id + " reached timeout, and no longer alive")

        elif((len(self.__sessions[session_id].client_list) == self.__sessions[session_id].session_capacity_max)): #Check list of clients, in relation to min capacity and max capacity
            self.__sessions[session_id].session_status = components._SESSION_ACTIVE  #if greater than min cap is met then session ready
            return
            # print("Session ready")
        
        
    def save_session_to_file(self,session_id,rootdir):
        if(os.path.isdir(rootdir+"/sessions") == False):
            os.makedirs(rootdir+"/sessions")
        file = open(rootdir+"/sessions/"+str(session_id) +".json",'w')
        session_stats = {"session_id" : session_id, 
                         "session_creation_time" : str(self.__sessions[session_id].session_creation_time),
                         "session_termination_time" : str(self.__sessions[session_id].session_termination_time),
                         "total_processing_time" : str(self.__sessions[session_id].total_processing_time),
                         "max_capacity" : self.__sessions[session_id].session_capacity_max,
                         "min_capacity" : self.__sessions[session_id].session_capacity_min,
                         "waiting_time" : str(self.__sessions[session_id].waiting_time),
                         "model_name" : self.__sessions[session_id].model_name,
                         "model_spec" : self.__sessions[session_id].model_spec,
                         "num_rounds" : self.__sessions[session_id].num_rounds,
                         "rounds" : []}
        for r in self.__sessions[session_id].rounds:
            r_str = {'participants' : r['participants'],
                        'status': r['status'], 
                        'acc':r['acc'],
                        'loss':r['loss'],
                        'starting_time' : str(r['starting_time']),
                        'completion_time' : str(r['completion_time']) ,
                        'processing_time': str(r['processing_time'])}
            session_stats['rounds'].append(r_str)
        json.dump(session_stats,file)
        print("Saved session stats into json file.")
            
    def get_session(self,session_id):
        self.update_session(session_id)
        return self.__sessions[session_id]
    
    def create_new_session(self,
                            session_id,
                            session_time,
                            session_capacity_min,
                            session_capacity_max, 
                            waiting_time, 
                            model_name,
                            model_spec,
                            fl_rounds):
        # try:
            new_session = Session(  session_id,
                                    session_time,
                                    session_capacity_min,
                                    session_capacity_max, 
                                    waiting_time, 
                                    model_name,
                                    model_spec,
                                    fl_rounds)
            
            self.__sessions[session_id] = new_session
            return 0
        # except:
        #     print("error occured in new session generation.")
        #     return -1

    def join_session(self,
                     session_id,
                     client_id,
                     client_role,
                     model_name,
                     model_spec,
                     fl_rounds,
                     memcap,
                     mdatasize,
                     pspeed):
        # try:
            if(session_id in self.__sessions):
                if(self.__sessions[session_id].num_rounds == int(fl_rounds)):
                    if(self.__sessions[session_id].model_name == model_name):
                        if(self.__sessions[session_id].model_spec == model_spec):
                            if(len(self.__sessions[session_id].client_list) < self.__sessions[session_id].session_capacity_max):
                                new_client = Client(client_id,
                                                    client_role,
                                                    fl_rounds,
                                                    memcap,
                                                    mdatasize,
                                                    pspeed)
                                self.__sessions[session_id].add_client(new_client)
                                return 0
                            else:
                                print("ERROR: session is full.")
                                return -1
                        else:
                            print("ERROR: model spec does not match")
                            # print(self.__sessions[session_id].num_rounds)
                            # print()
                            return -2
                    else:
                        print("ERROR: model name does not match")
                        return -3
                else:
                    print("ERROR: number of FL rounds does not match.")
                    return -4
            else:
                print("ERROR: session id does not excist")
                return -5
        # except:
        #     print("error occured in joining client to session.")
        #     return -5

    def plot_accloss(self,acc,loss, rounds = 0, init = False):
      
        if(init == True): 
            self.plot_fig, (self.plot_ax_accloss, self.plot_ax_mem,self.plot_ax_total_mem) = plt.subplots(3,1,layout='constrained') # fig : figure object, ax : Axes object
            self.plot_ax_accloss.set_xlabel('round')
            self.plot_ax_accloss.set_xlim((0,rounds))
            
            self.plot_ax_accloss.set_ylim((0.0,1.0))
            self.plot_ax_accloss.set_ylabel('prediction accuracy')

            self.plot_ax_mem.set_xlim((0,rounds))
            self.plot_ax_mem.set_ylabel('ram usage (bytes)')
            self.plot_ax_mem.set_xlabel('round')

            self.plot_ax_total_mem.set_xlim((0,rounds))
            self.plot_ax_total_mem.set_ylabel('total ram usage (bytes)')
            self.plot_ax_total_mem.set_xlabel('round')
            
            self.plt_step = []
            self.plt_acc = []
            self.plt_loss = []
        else:
            self.plt_step.append(len(self.plt_step))
            self.plt_acc.append(float(acc))
            self.plt_loss.append(float(loss))
           
        # plt.plot(self.plt_step,self.plt_loss,color='red')

        self.plot_ax_accloss.plot(self.plt_step,self.plt_loss,color='red')
        self.plot_ax_accloss.plot(self.plt_step,self.plt_acc,color='blue')
        
        for cl in self.mem_usage_track:
            self.plot_ax_mem.plot(self.plt_step,self.mem_usage_track[cl])

        for cl in self.total_mem_usage:
            self.plot_ax_total_mem.plot(self.plt_step,self.total_mem_usage[cl])
       
        plt.pause(0.1)
        
        # plt.show()



import json
import torch
import torch.nn as nn
# from Global.custom_models import VGG, MNISTMLP
# from Global.custom_datasets import CIFAR10, MNIST
from Base.executable_class import PubSub_Base_Executable
from io import BytesIO
import zlib

import base64

class dflmq_parameter_server(PubSub_Base_Executable):
    
    def __init__(self , 
                 myID : str , 
                 broker_ip : str , 
                 broker_port : int , 
                 introduction_topic : str , 
                 controller_executable_topic : str , 
                 controller_echo_topic : str ,
                 start_loop : bool) -> None : 
        
       
        self.CoTPST = "Coo_to_PS_T"
        self.PSTCoT = "PS_to_Coo_T"
        self.PSTCliT = "PS_to_Cli_T"
        self.PSTCliIDT = "PS_to_Cli_ID_"
        self.executables.append('broadcast_model')
        self.executables.append('publish_dataset')
        
        super().__init__(
                    myID , 
                    broker_ip , 
                    broker_port , 
                    introduction_topic , 
                    controller_executable_topic , 
                    controller_echo_topic , 
                    start_loop)

        # self.model_stash   =   {'MNISTMLP' : MNISTMLP(),
        #                         'VGG11' : VGG('VGG11'),
        #                         'VGG3' : VGG('VGG3')}
        # self.dataset_stash        =   {'MNIST' : MNIST(),
        #                                'CIFAR10' : CIFAR10()}

        self.client.subscribe(self.CoTPST)

    def __execute_on_msg(self,header_parts, body):
        super().__execute_on_msg(header_parts, body)
        # header_parts = self._get_header_body(msg)
        if header_parts[2] == 'broadcast_model' :
            model_name = body.split('-model_name ')[1].split(';')[0]
            self.broadcast_model(model_name)
            
        if header_parts[2] == 'publish_dataset':
            dataset_name = body.split('-dataset_name ')[1].split(' -num_clients ')[0]
            num_clients = int(body.split('-num_clients ')[1].split(' -ids ')[0])
            ids = (body.split(' -ids ')[1].split(';')[0]).split(' ')
            if(num_clients != len(ids)):
                print("number of clients does not match with number of ids passed.")
            else:
                self.publish_dataset(num_clients, dataset_name, ids)
      
    def broadcast_model(self,model_name):
        model = self.model_stash[model_name]
        weights_and_biases = {}
        for name, param in model.named_parameters():
            weights_and_biases[name] = param.data.tolist()

        model_params = json.dumps(weights_and_biases)
        print(len(model_params))
        self.publish(self.PSTCliT,"construct_logic_model"," -id all -model_name " + model_name + " -model_params " + str(model_params)) 
        
    def publish_dataset(self, num_clients, dataset_name, client_ids):
        
        [traindata_splits, testdata] = self.dataset_stash[dataset_name].load_data_for_clients(num_clients)
        for i in range(num_clients):
            print("buffering training dataset for client " + str(client_ids[i]))
            buffer = BytesIO()
            torch.save({'trainset': traindata_splits[i], 'testset': testdata}, buffer)
            buffer.seek(0)  # Rewind the buffer to the beginning
            
            #bin_dataset = buffer.read()
            #bin_datasets = f'{buffer.read()}'
            compressed_data = zlib.compress(buffer.read())
            compressed_data_s = base64.b64encode(compressed_data).decode('utf-8')

            self.publish(self.PSTCliIDT + client_ids[i],"collect_logic_data"," -id " + client_ids[i] + " -dataset_name " + dataset_name + " -dataset_type " + "training,testing"  + " -data " + compressed_data_s) 
            buffer.close()


# userID = input("Parameter Server ID: ")
# print("PS with ID=" + userID +" is created.")

# exec_program = dflmq_parameter_server(myID = userID,
#         broker_ip = 'localhost' ,
#         broker_port = 1883,
#         introduction_topic='client_introduction',
#         controller_executable_topic='controller_executable',
#         controller_echo_topic="echo",
#         start_loop=False
# )
# exec_program.base_loop()




import json
import numpy
import torch
import torch.nn as nn

class Model_Controller():
    def __init__(self,client_id):
        self.client_id = client_id
        self.Models = {}

    def get_model(self,session_id):
        return self.Models[session_id]
    
    def set_model(self, session_id,model):
        self.Models[session_id] = model

    def update_model(self,session_id,model_params):
        keys = list(model_params.keys())
        for k in keys:
            model_params[k] = torch.tensor(model_params[k])
        self.Models[session_id].load_state_dict(model_params)
    
    def get_model_spec(self,model,ml_framework = "pytorch"):
        if(ml_framework == "pytorch"):
            model_info = self.get_pytorch_model_info(model)
            return json.dumps(model_info, indent=4)
        return ""
    
    def get_pytorch_model_info(self,model):
        model_info = []
        for name, layer in model.named_children():  # Iterate only through direct child layers
            layer_info = {
                "layer_name": name,
                "layer_type": layer.__class__.__name__,
                "num_parameters": sum(p.numel() for p in layer.parameters() if p.requires_grad),
            }
            model_info.append(layer_info)

        return model_info




import tkinter as tk
import time
from .modules import controller_core_modules

class Controller_Dashboard:
    ##__________________________________________MQTT___________________________________

    def __init__(self):

        self.root = tk.Tk()    
        self.server = controller_core_modules.Server_Controller(log_func=self.Print_onShellOutput)
        self.server.MQTT_thread_handler.setDaemon(True)

        self.command_history = []
        self.command_history_curr_index = 0
        self.selected_client = 0
        self.init_menubar()
        self.init_tkinter_elements()
        self.init_Lists()

        self.shell_output.config(state=tk.DISABLED)
        
    #_____________________________________________FUNCTIONS____________________________

    def on_connect(self,client,userdata, flags, rc):
        print("Connect")
        self.Print_onShellOutput("Connected to broker " + self.server.selected_broker[0] + " with result code " + str(rc))
        self.server.Subscribe_to_All()

    def on_message(self,client,userdata, msg):
        raw_msg = str(msg.payload.decode())
        self.Print_onShellOutput("CLIENT:: " + raw_msg)
        self.server.message_parse(raw_msg)
        self.refresh_executables_list()
        self.refresh_client_list()

    def Connect_to_Broker(self):
        self.server.Connect_to_Selected_Broker(self.on_connect,self.on_message) 

    def Clients_Introduce(self):
        
        self.refresh_client_list()
        self.server.Pub_Clients_Introduce()

    def Client_PowerOn(self):
        print(self.selected_client)
        self.server.Client_PowerOn(self.selected_client,self.refresh_client_list)
    
    def Client_PowerOff(self):
        self.server.Client_PowerOff(self.selected_client,self.refresh_client_list)

    def Print_onShellOutput(self,msg):
        print(msg)
        self.shell_output.config(state= tk.NORMAL)
        self.shell_output.insert(tk.END, "\n" + msg )
        self.shell_output.see(tk.END)
        self.shell_output.config(state= tk.DISABLED)

    def CheckEntry(self,e):
        #self.Print_onShellOutput("CONTROLLER::INPUT>> "  + self.userInput.get())
        self.command_history.append(self.userInput.get())
        self.server.Command_Parser(self.userInput.get())
        
        self.command_history_curr_index = len(self.command_history) - 1 
        self.userInput.delete(0,'end')

    def Navigate_cmdHist_down(self,e):

        if(len(self.command_history) == 0):
            return
        
        if( self.command_history_curr_index > 0):
            self.command_history_curr_index -= 1
        
        self.userInput.delete(0,'end')
        self.userInput.insert(0,self.command_history[self.command_history_curr_index])


    def Navigate_cmdHist_up(self,e):
        
        if(len(self.command_history) == 0):
            return
        
        if( self.command_history_curr_index < len(self.command_history) - 1):
            self.command_history_curr_index += 1
        
        self.userInput.delete(0,'end')
        self.userInput.insert(0,self.command_history[self.command_history_curr_index])

    def donothing():
        x = 0

    def print_hello():
        print("Hello")

    
    def refresh_client_list(self):
        self.Lb1.delete(0,tk.END)
        cc = 0
        for item in self.server.client_list:
            self.Lb1.insert(cc, item.name)
            cc += 1

    def refresh_executables_list(self):
        self.Lb4.config(state= tk.NORMAL)
        self.Lb4.delete('1.0',tk.END)
        self.Lb4.insert(tk.END,"Client Executables: \n\n")
        self.Lb4.insert(tk.END,str(self.server.registered_executables))
        self.Lb4.insert(tk.END,"\n______________\n\nController Procedures: \n\n")
        self.Lb4.insert(tk.END,str(self.server.my_procedures.registered_procedures))
        self.Lb4.see(tk.END)
        self.Lb4.config(state= tk.DISABLED)
    ##____________________________________________LIST SELECT FUNCTIONS _______________

    def broker_list_function(self,e):
        self.server.selected_broker = self.server.broker_list[self.Lb2.curselection()[0]]
        print("Selected Broker now is: " + self.server.selected_broker[0])
    
    def client_list_function(self,e):
        self.selected_client = self.Lb1.curselection()[0]
        self.refresh_client_properties()
    
    def refresh_client_properties(self):
        self.Lb3.config(state=tk.NORMAL)
        self.Lb3.delete('1.0',tk.END)
        self.Lb3.insert(tk.END,"Client Name = " + self.server.client_list[self.selected_client].name + "\n")
        self.Lb3.insert(tk.END,"Client ID = " + self.server.client_list[self.selected_client].id + "\n")
        self.Lb3.insert(tk.END,"Client broker = " + self.server.client_list[self.selected_client].broker + "\n")
        self.Lb3.insert(tk.END,"Client Executables = " + str(self.server.client_list[self.selected_client].executables) + "\n")
    
    ##____________________________________________INITIALIZERS__________________________

    def init_Lists(self):
        #CLIENT LIST
        #BROKER LIST
            bc = 0
            for item in self.server.broker_list:
                self.Lb2.insert(bc, item[0])
                bc += 1

        #CLIENT PROPERTIES
            self.Lb3.config(state= tk.DISABLED)

        #EXECUTABLE LIST
            self.Lb4.config(state= tk.DISABLED)

    def init_menubar(self):

        self.menubar = tk.Menu(self.root)
        filemenu = tk.Menu(self.menubar, tearoff=0)
        filemenu.add_command(label="New", command=self.donothing)
        filemenu.add_command(label="Open", command=self.donothing)
        filemenu.add_command(label="Save", command=self.donothing)
        filemenu.add_separator()
        filemenu.add_command(label="Exit", command=self.root.quit)
        self.menubar.add_cascade(label="File", menu=filemenu)
        helpmenu = tk.Menu(self.menubar, tearoff=0)
        helpmenu.add_command(label="Help Index", command=self.donothing)
        helpmenu.add_command(label="About...", command=self.donothing)
        self.menubar.add_cascade(label="Help", menu=helpmenu)
        self.root.config(menu=self.menubar)


    def init_tkinter_elements(self):

        self.right_frame = tk.Frame(self.root, bd=2)
        self.left_frame = tk.Frame(self.root, bd=2)
        self.inner_right_top_frame = tk.Frame(self.right_frame, bd=3)
        self.inner_right_bottom_frame = tk.Frame(self.right_frame, bd=2)

        self.action_buttons_frame = tk.Frame(self.right_frame, bd=10)

        self.Lb1 = tk.Listbox(self.left_frame, justify="left",background="#26242f",foreground="white")
        #Lb1.insert(1, "list1")
        self.Lb2 = tk.Listbox(self.left_frame, justify="left",background="#26242f",foreground="white")
        self.Lb3 = tk.Text(self.inner_right_top_frame,background="#26242f",foreground="white",height=2,width=1,wrap=tk.WORD)
        #Lb3.insert(1, "list3")
        self.Lb4 = tk.Text(self.inner_right_top_frame,background="#26242f",foreground="white",height=2,width=1,wrap=tk.WORD)
        #Lb4.insert(1, "list4")

        self.userInput = tk.Entry(self.inner_right_bottom_frame,background="#26244c",foreground="white")
        self.shell_output = tk.Text(self.inner_right_bottom_frame,background="#26242f",foreground="white",height=2,width=1)

        self.ac_bt0 = tk.Button(self.action_buttons_frame,text="Power ON")
        self.ac_bt1 = tk.Button(self.action_buttons_frame,text="Power OFF")
        self.ac_bt2 = tk.Button(self.action_buttons_frame,text="3"    ,command=self.print_hello)
        self.ac_bt3 = tk.Button(self.action_buttons_frame,text="4"    ,command=self.print_hello)
        self.ac_bt4 = tk.Button(self.action_buttons_frame,text="5"    ,command=self.print_hello)
        self.ac_bt5 = tk.Button(self.action_buttons_frame,text="6"    ,command=self.print_hello)
        self.ac_bt6 = tk.Button(self.action_buttons_frame,text="7"    ,command=self.print_hello)
        self.ac_bt7 = tk.Button(self.action_buttons_frame,text="8"    ,command=self.print_hello)
        self.ac_bt8 = tk.Button(self.action_buttons_frame,text="Clients Introduce")
        self.ac_bt9 = tk.Button(self.action_buttons_frame,text="Connect to Broker")

        self.userInput.bind("<Return>",self.CheckEntry)
        self.userInput.bind("<Up>",self.Navigate_cmdHist_down)
        self.userInput.bind("<Down>",self.Navigate_cmdHist_up)
        self.Lb2.bind("<<ListboxSelect>>", self.broker_list_function)
        self.Lb1.bind("<<ListboxSelect>>", self.client_list_function)
        
        self.root.title("MQTT Fleet Controller Dashboard")
        self.root.config(bg="#26242f")   
        self.root.geometry("1440x800")

        self.rootHeight = self.root.winfo_height()
        self.rootWidth =  self.root.winfo_width()

        self.root.rowconfigure(0, weight=1)
        # root.rowconfigure(1, weight=1)
        self.root.columnconfigure(0, weight=1)
        self.root.columnconfigure(1, weight=4)

        self.right_frame.columnconfigure(0,weight=1)
        self.right_frame.rowconfigure(0)
        self.right_frame.rowconfigure(1,weight=10)
        self.right_frame.rowconfigure(2,weight=10)

        self.left_frame.columnconfigure(0,weight=1)
        self.left_frame.rowconfigure(0,weight=2)
        self.left_frame.rowconfigure(1,weight=1)

        self.inner_right_top_frame.columnconfigure(0,weight=1)
        self.inner_right_top_frame.columnconfigure(1,weight=1)
        self.inner_right_top_frame.rowconfigure(0,weight=1)

        self.inner_right_bottom_frame.columnconfigure(0,weight=1)
        self.inner_right_bottom_frame.rowconfigure(0,weight=7)
        self.inner_right_bottom_frame.rowconfigure(1,weight=1)

        self.action_buttons_frame.rowconfigure(0,weight=1)
        self.action_buttons_frame.columnconfigure(0,weight=1)
        self.action_buttons_frame.columnconfigure(1,weight=1)
        self.action_buttons_frame.columnconfigure(2,weight=1)
        self.action_buttons_frame.columnconfigure(3,weight=1)
        self.action_buttons_frame.columnconfigure(4,weight=1)
        self.action_buttons_frame.columnconfigure(5,weight=1)
        self.action_buttons_frame.columnconfigure(6,weight=1)
        self.action_buttons_frame.columnconfigure(7,weight=1)
        self.action_buttons_frame.columnconfigure(8,weight=1)
        self.action_buttons_frame.columnconfigure(9,weight=1)

        self.right_frame                 .grid(row=0,column=1,sticky='nwse',padx=2,pady=2)
        self.left_frame                  .grid(row=0,column=0,sticky='nwse',padx=2,pady=2)
        self.action_buttons_frame        .grid(row=0,column=0,sticky='nwse',padx=1,pady=1)
        self.inner_right_top_frame       .grid(row=1,column=0,sticky='nwse',padx=1,pady=1)
        self.inner_right_bottom_frame    .grid(row=2,column=0,sticky='nwse',padx=1,pady=1)
        self.Lb1.grid(row=0,column=0,sticky='nwse',padx=1,pady=1)
        self.Lb2.grid(row=1,column=0,sticky='nwse',padx=1,pady=1)
        self.Lb3.grid(row=0,column=0,sticky='nwse',padx=1,pady=1)
        self.Lb4.grid(row=0,column=1,sticky='nwse',padx=1,pady=1)
        self.shell_output.grid(row=0,column=0,sticky='nwse',padx=2,pady=2)
        self.userInput.grid(row=1,column=0,sticky='nwse',padx=2,pady=2)
        #self.ac_bt0.grid(row=0,column=0,sticky='nswe')
        #self.ac_bt1.grid(row=0,column=1,sticky='nswe')
        # self.ac_bt2.grid(row=0,column=2,sticky='nswe')
        # self.ac_bt3.grid(row=0,column=3,sticky='nswe')
        # self.ac_bt4.grid(row=0,column=4,sticky='nswe')
        # self.ac_bt5.grid(row=0,column=5,sticky='nswe')
        # self.ac_bt6.grid(row=0,column=6,sticky='nswe')
        # self.ac_bt7.grid(row=0,column=7,sticky='nswe')
        self.ac_bt8.grid(row=0,column=8,sticky='nswe')
        self.ac_bt9.grid(row=0,column=9,sticky='nswe')

        self.ac_bt0.config(command=self.Client_PowerOn)
        self.ac_bt1.config(command=self.Client_PowerOff)
        self.ac_bt9.config(command=self.Connect_to_Broker)
        self.ac_bt8.config(command=self.Clients_Introduce)        

    
def main():
    dashbooard = Controller_Dashboard()
    dashbooard.root.mainloop()

#____________________________________________GUI START_____________________________________________

if __name__ == "__main__":
    main()
	



import paho.mqtt.client as mqtt
import time as T
import os
class MQTTFC_Procedures:

    def __init__(self):
        self.wait_time = .5 # in sec
        self.registered_procedures = ['Topic_Naming_PUF_SinglePair']
    
    def parse_procedure_command(self,command, cmdline,fc_parser_func):
        if(command in self.registered_procedures):
            if(command == 'Topic_Naming_PUF_SinglePair'):

                id1 = cmdline.split(' -id1 ')[1].split(' -id2 ')[0]
                id2 = cmdline.split(' -id2 ')[1].split(';')[0]
                self.Topic_Naming_PUF_SinglePair(id1,id2,fc_parser_func)

    def Topic_Naming_PUF_SinglePair(self,id1,id2, parser_func):
        
        rep = 1000
        topic_sizes = [16,32,64,96,128,192]
        code_length = [1,2,3,4,5,6,7,8,9,10]
        
        # topic_sizes = [16]
        # code_length = [4]
        
        for ts in topic_sizes:
            for cl in code_length:
                for i in range(rep):
                    parser_func("run establish_private_topic -id1 " + str(id1) + " -id2 " + str(id2) + " -l " + str(ts) + " -rl " + str(cl))
                    T.sleep(self.wait_time)
                    parser_func("run wisper -id1 " + str(id1) + " -id2 " + str(id2))
                    T.sleep(self.wait_time)


import json
import psutil
import os
import psutil

class Role_Arbiter():
    def __init__(self,client_id):

        self.is_aggregator = False
        self.is_root_aggregator = False
        
        self.client_id = client_id
        self.current_session = ""
        self.roles = {} #Keys are session ids
        self.cluster_heads = {}
        self.session_role_dicionaries = {}
    def get_resources(self):
        resources = {
            'cpu_count'     : psutil.cpu_count() ,
            'cpu_frequency' : psutil.cpu_freq() ,
            'ram_usage'     : psutil.virtual_memory()[3]/1000000000 ,
            # 'net_counters'  : psutil.net_io_counters()
            # 'cpu_stats'     : psutil.cpu_stats() ,
            # 'net_stats'     : psutil.net_if_stats() ,
            # 'disk_usage'    : psutil.disk_usage("/") ,
            }

        res_msg = json.dumps(resources)
        return res_msg
    
    def get_role(self,session_id):
        return self.roles[session_id]

    def set_current_session(self,session_id):
        self.current_session = session_id
        
    def add_session(self,session_id,):
        self.roles[session_id] = ""

    def set_role(self,session_id,role):
        if(session_id in self.roles):
            self.roles[session_id] = role
            
            if(role == "agg_0_" + str(session_id)):
                self.is_aggregator = True
                self.is_root_aggregator = True
                
            elif(role[0] == 'a'):
                self.is_aggregator = True
                self.is_root_aggregator = False
            elif(role[0] == 't'):
                self.is_aggregator = False
                self.is_root_aggregator = False
            else:
                print("Unknown role!!")
                return -1
            
            role_dic = self.session_role_dicionaries[session_id]
            agg_clients = role_dic.keys()
            for agg_c in agg_clients:
                sub_clients = role_dic[agg_c]
                if(role in sub_clients):
                    print("Found matching role in role dictionary. Returning aggregator of the cluster")
                    self.cluster_heads[session_id] = agg_c
                    # return agg_c

            return 0
            
        else:
            print("No session found with session id: " + str(session_id))
            
    def reset_role(self,session_id,role):
        if(session_id in self.roles):
            self.roles[session_id] = role
            if(role == "agg_0_" + str(session_id)):
                self.is_aggregator = True
                self.is_root_aggregator = True
                
            elif(role[0] == 'a'):
                self.is_aggregator = True
                self.is_root_aggregator = False
            elif(role[0] == 't'):
                self.is_aggregator = False
                self.is_root_aggregator = False
            else:
                print("Unknown role!!")
                return -1
            
            role_dic = self.session_role_dicionaries[session_id]
            agg_clients = role_dic.keys()
            for agg_c in agg_clients:
                sub_clients = role_dic[agg_c]
                if(role in sub_clients):
                    print("Found matching role in role dictionary. Returning aggregator of the cluster")
                    self.cluster_heads[session_id] = agg_c
                    # return agg_c
            return 0
        else:
            print("No session found with session id: " + str(session_id))

    def set_role_dicionary(self,session_id,roles):
        self.session_role_dicionaries[session_id] = json.loads(roles)
    
    def get_session_aggregator(self,session_id):
        return self.cluster_heads[session_id]



import numpy as np
import torch


class SDFLMQ_Aggregator():
    
    def __init__(self)-> None:
        self.max_clients = {}
        self.client_model_params = {}

    def set_max_agg_capacity(self,session_id,max_clients):
        self.max_clients[session_id] = max_clients
    
    def accumulate_params(self,session_id, local_model, params):
        if(session_id not in self.client_model_params):
            self.client_model_params[session_id] = [params]    
        else:
            self.client_model_params[session_id].append(params)
        
        if(len(self.client_model_params[session_id]) >= self.max_clients[session_id]):
            print("number of contributing clients: " + str(len(self.client_model_params[session_id])))
            g_model = self.fed_average(session_id,local_model)
            return [1,g_model]
        else:
            return [0,None]
        
    def fed_average(self,session_id, local_model):
      
        global_dict = {}
        for name, param in local_model.named_parameters():
            global_dict[name] = param.data.tolist()

        # print("Length of client model params list: " + str(len(self.client_model_params)))
        num_clients = len(self.client_model_params[session_id])
        param_names = global_dict.keys()
        for name in param_names:
            for i in range(num_clients):
                # print(global_dict[name])
                # print(self.client_model_params[i][name])
                if(name in self.client_model_params[session_id][i]):
                    layer = self.client_model_params[session_id][i][name]
                    global_dict[name] = torch.tensor(global_dict[name]) + torch.tensor(layer) 
                    # print(global_dict[name])
            global_dict[name] =  global_dict[name] / num_clients
        local_model.load_state_dict(global_dict)
        self.client_model_params[session_id] = []
        return global_dict




from random import randint , random , sample , seed 
from os import system
import copy 


# Particle class
class Particle :
    def __init__(self, pos , fitness , velocity , best_pos_fitness) : 
        self.pos = pos
        self.fitness = fitness
        self.velocity = velocity
        self.best_pos = self.pos.copy()
        self.best_pos_fitness = best_pos_fitness

# Swarm class
class Swarm : 
    def __init__(self , pop_n , particle_size, num_clients, random_poses = []) :
        
        self.particles = self.__generate_random_particles(pop_n , particle_size,num_clients)
        for i in range(len(random_poses)):
            self.particles[i].pos = random_poses[i]
            self.particles[i].best_pos = random_poses[i]
            
        # self.global_best_particle = copy.deepcopy(max(self.particles, key=lambda particle: particle.fitness))
        self.global_best_particle = copy.deepcopy(self.particles[0])

    def __generate_random_particles(self, pop_n, particle_size, num_clients):
        init_particle_pos = [randint(0,num_clients) for _ in range(particle_size)]
        particles = []

        for i in range(pop_n):
            particle_pos = init_particle_pos
            fitness = None
            velocity = [0 for _ in range(particle_size)]       
            best_pos_fitness = None                   
            particles.append(Particle(particle_pos, fitness, velocity, best_pos_fitness))

        return particles


class PSO:
    def __init__(self, particle_num, particle_size, num_clients, random_poses):

        self.iw = .1                                     # Inertia Weight (Higher => Exploration | Lower => Exploitation)   (0.1 , 0.5)
        self.c1 = .1                                     # Pbest coefficient (0.01 , 0.1)
        self.c2 = 1                                       # Gbest coefficient 
        self.pop_n = particle_num                                    # Population number (3 , 5 , 10 , 15 , 20*)
        self.max_iter = 1000  
        self.particle_size = particle_size

        # System parameters
        # self.randomness_seed = 10
        self.tracking_mode = False   
        self.velocity_factor = 0.15                       # Increasing velocity_factor causes more exploration resulting higher fluctuations in the particles plot (default range between 0 and 1 (Guess))

        self.particle_counter = -1
        self.iter_counter = -1

        self.swarm = Swarm(self.pop_n , self.particle_size,num_clients,random_poses)
        # if self.tracking_mode : 
        #     seed(self.randomness_seed)

        
        # Experiment parameters
        # stepper = 1
        # scenario_file_number = randint(1,1000)
        # scenario_folder_number = randint(1,1000)                       
        # scenario_folder_name = f"scenario_case_{scenario_folder_number}"

        # # Graph illustration required parameters 
        # particles_fitness_fig_path = f"./measurements/results/{scenario_folder_name}/particles_fitness_{scenario_file_number}.pdf"
        # swarm_best_fitness_fig_path = f"./measurements/results/{scenario_folder_name}/swarm_best_fitness_{scenario_file_number}.pdf"
        # tpd_fig_path = f"./measurements/results/{scenario_folder_name}/tpd_{scenario_file_number}.pdf"

        # sbpfl = ("iteration" , "best particle fitness")
        # pfl = ("iteration" , "particles fitness") 
        # tpdl = ("iteration" , "total processing delay")

        # sbpft = "Swarm’s Best Particle Fitness Plot"
        # pft = "Particles Fitness Plot"
        # tpdt = "Total Processing Delay Plot"

        self.gbest_particle_fitness_results = []
        self.particles_fitnesses_buffer = []
        self.particles_fitnesses_tuples = []

        self.tpd_buffer = []
        self.tpd_tuples = []
        self.iterations = []

        # # CSV output required parameters
        # csv_particles_output_file_name = f"particles_data_{scenario_file_number}"
        # csv_swarm_best_output_file_name = f"swarm_best_data_{scenario_file_number}"
        # csv_tpd_output_file_name = f"tpd_data_{scenario_file_number}"

        # csv_particles_data_path = f"./measurements/results/{scenario_folder_name}/{csv_particles_output_file_name}.csv"
        # csv_swarm_best_data_path = f"./measurements/results/{scenario_folder_name}/{csv_swarm_best_output_file_name}.csv"
        # csv_tpd_data_path = f"./measurements/results/{scenario_folder_name}/{csv_tpd_output_file_name}.csv"

        # particles_columns = ["iteration"] + [f"particle_{i+1}_fitness" for i in range(pop_n)]
        # swarm_best_columns = ["iteration", "swarm_best_fitness"]
        # tpd_columns = ["iteration"] + [f"tpd_particle_{i+1}" for i in range(pop_n)]

        # csv_cols = [particles_columns, swarm_best_columns, tpd_columns]
        # csv_rows = [[], [], []]

        # JSON output required parameters (Particles constant metadata)
        # json_path = f"./measurements/results/{scenario_folder_name}/pso_scenario_case_{scenario_file_number}.json"
        # json_pso_dict = {
        #     "DEPTH" : DEPTH,
        #     "WIDTH" : WIDTH,
        #     "dimensions" : dimensions,
        #     "randomness_seed" : randomness_seed,
        #     "iw" : iw,
        #     "c1" : c1,
        #     "c2" : c2,
        #     "pop_n" : pop_n,
        #     "max_iter" : max_iter,
        #     "velocity_factor" : velocity_factor
        # } 


    # Fitness function
    def processing_fitness(self,tpd):
        return -1 * float(tpd)

    def updateVelocity(self,current_velocity, current_position, personal_best, global_best, iw, c1, c2):
        r1 = [random() for _ in range(len(current_velocity))]
        r2 = [random() for _ in range(len(current_velocity))]

        inertia = [iw * v for v in current_velocity]
        cognitive = [c1 * r1[i] * (personal_best[i] - current_position[i]) for i in range(len(current_velocity))]
        social = [c2 * r2[i] * (global_best[i] - current_position[i]) for i in range(len(current_velocity))]
        
        max_velocity = max(1, int(len(current_velocity) * self.velocity_factor))
        new_velocity = [round(inertia[i] + cognitive[i] + social[i]) for i in range(len(current_velocity))]
        new_velocity = [max(min(v, max_velocity), -max_velocity) for v in new_velocity]  # Apply velocity limits

        return new_velocity

    def applyVelocity(self,p_position, p_velocity):
        new_position = []
        client_count = len(p_position)

        for a, b in zip(p_position, p_velocity):
            np = (a + b) % client_count  

            while np in new_position:
                np = (np + 1) % client_count  

            new_position.append(np)

        return new_position

    def get_next_particle(self):
        self.particle_counter += 1
        
        if(self.particle_counter >= len(self.swarm.particles)):
            self.iter_counter += 1
            self.particle_counter = 0

        print("particle index: " + str(self.particle_counter))
        print("len swarm particles: " + str(len(self.swarm.particles)))
        next_particle = copy.copy(self.swarm.particles[self.particle_counter].pos)

        return next_particle

    def optimize(self,total_processing_delay):    

        if(self.swarm.particles[self.particle_counter].best_pos_fitness == None or 
           self.swarm.particles[self.particle_counter].fitness == None):
            new_fitness = self.processing_fitness(total_processing_delay)
            self.swarm.particles[self.particle_counter].fitness = copy.copy(new_fitness)
            self.swarm.particles[self.particle_counter].best_pos_fitness = copy.copy(new_fitness)
            if(self.particle_counter == 0):
                self.swarm.global_best_particle = copy.deepcopy(self.swarm.particles[self.particle_counter])
            elif(self.swarm.particles[self.particle_counter].fitness > self.swarm.global_best_particle.fitness):
                self.swarm.global_best_particle = copy.deepcopy(self.swarm.particles[self.particle_counter]) 

            return

        new_velocity = self.updateVelocity(self.swarm.particles[self.particle_counter].velocity,
                                            self.swarm.particles[self.particle_counter].pos, 
                                            self.swarm.particles[self.particle_counter].best_pos, 
                                            self.swarm.global_best_particle.best_pos, self.iw, self.c1, self.c2)
        new_position = self.applyVelocity(self.swarm.particles[self.particle_counter].pos, new_velocity)

        new_pos_fitness = self.processing_fitness(total_processing_delay)

        self.swarm.particles[self.particle_counter].pos = new_position
        self.swarm.particles[self.particle_counter].fitness = new_pos_fitness
        self.swarm.particles[self.particle_counter].velocity = new_velocity
        
        if self.swarm.particles[self.particle_counter].fitness > self.swarm.particles[self.particle_counter].best_pos_fitness :  
            self.swarm.particles[self.particle_counter].best_pos = self.swarm.particles[self.particle_counter].pos.copy()
            self.swarm.particles[self.particle_counter].best_pos_fitness = copy.copy(self.swarm.particles[self.particle_counter].fitness)

            if self.swarm.particles[self.particle_counter].fitness > self.swarm.global_best_particle.fitness:
                self.swarm.global_best_particle = copy.deepcopy(self.swarm.particles[self.particle_counter])              
        
        self.tpd_buffer.append(total_processing_delay)

        self.particles_fitnesses_buffer.append(self.swarm.particles[self.particle_counter].fitness)

        
        #root = reArrangeHierarchy(new_position)
        
        

            # iw = 0.15 - ((0.14 * counter) / max_iter) 

            # iterations.append(counter)
            
            # gbest_particle_fitness_results.append(swarm.global_best_particle.fitness)
            # tpd_tuples.append(tpd_buffer.copy())
            # particles_fitnesses_tuples.append(particles_fitnesses_buffer.copy()) # We could simply reverse the TPD plot and get Particles Fitnesses Plot but as the fitness function might change later this method is not reliable 
            
            # particles_row = [counter] + [round(fitness , 2) for fitness in particles_fitnesses_buffer]
            # csv_rows[0].append(particles_row)
            
            # swarm_best_row = [counter, round(swarm.global_best_particle.fitness , 2)]
            # csv_rows[1].append(swarm_best_row)
            
            # tpd_row = [counter] + [round(tpd , 2) for tpd in tpd_buffer]
            # csv_rows[2].append(tpd_row)


            # system("clear")
        
            # print("Iteration : " , counter)
            
            # tpd_buffer.clear()
            # particles_fitnesses_buffer.clear()
            


        # save_data_to_csv(csv_cols[0] , csv_rows[0] , csv_particles_data_path)
        # save_data_to_csv(csv_cols[1] , csv_rows[1] , csv_swarm_best_data_path)
        # save_data_to_csv(csv_cols[2] , csv_rows[2] , csv_tpd_data_path)
        # save_metadata_to_json(json_pso_dict , json_path)

        # illustrate_plot(gbest_particle_fitness_results , sbpfl , sbpft , swarm_best_fitness_fig_path)
        # plot_tuple_curves(particles_fitnesses_tuples , pfl , pft , particles_fitness_fig_path)
        # plot_tuple_curves(tpd_tuples , tpdl , tpdt , tpd_fig_path)
        


#Here, the core modules of MQTT Fleet Control is placed
#import broker_manager
from .core import clients_manager
#import executable_manager
import paho.mqtt.client as mqtt
import os
import threading
import json
import time as T
import importlib.resources as pkg_resources
from .core.MQTTFC_Procedures import MQTTFC_Procedures
from . import metadata

class Server_Controller:
    
    
    sub_topics_directory =  os.path.dirname(os.path.abspath(__file__)) + "/metadata/topics_to_subscribe.json"
    pub_topics_directory =  os.path.dirname(os.path.abspath(__file__)) + "/metadata/topics_to_publish.json"
    broker_mtf_directory =  os.path.dirname(os.path.abspath(__file__)) + "/metadata/broker_metadata.json"
    client_mtf_directory =  os.path.dirname(os.path.abspath(__file__)) + "/metadata/client_metadata.json"
    
    registered_executables = []
    
    client = mqtt.Client(mqtt.CallbackAPIVersion.VERSION1)
    MQTT_thread_handler = threading.Thread(target=client.loop_forever)
    
    def __init__(self,log_func):

        self.broker_list = self.Load_brokers()
        self.selected_broker = self.broker_list[0]
        self.my_procedures = MQTTFC_Procedures()

        #self.client_list = self.Load_clients()
        self.client_list = []

        self.sub_topics = self.load_topics_to_subscribe()
        self.pub_topics = self.load_topics_to_publish()
        
        if(log_func != None):
            self.log = log_func
        else:
            self.log = self.Log

    def Subscribe_to_All(self):
        for topic in self.sub_topics:
            self.Subscribe(topic_name=topic)

    def Subscribe(self,topic_name):
        self.client.subscribe(topic=topic_name)
        print("Subscribed to topic " + topic_name)
        self.log ("CONTROLLER::CORE:: " + "Subscribed to topic " + topic_name)
    
    def Connect_to_Selected_Broker(self,on_connect_func, on_message_func):
        try:
            if(not self.MQTT_thread_handler.is_alive()):
                self.client.on_connect = on_connect_func
                self.client.on_message = on_message_func
        
                self.log("CONTROLLER::CORE:: connecting to " + str(self.selected_broker))
                
                self.client.connect(self.selected_broker[0],self.selected_broker[1],self.selected_broker[2])
                self.MQTT_thread_handler.start()

            else:
                self.log("CONTROLLER::CORE:: Dashboard already connected to a broker. Close the dashboard and start again.")
        
        except:
                self.log("CONTROLLER::CORE:: Error in connecting to the broker. Or no broker is selected.")
            

    def Client_PowerOn(self,c_index,callback):
        return 0
            
    def Client_PowerOff(self,c_index,callback):
        return 0
        
    def Load_brokers(self):
        brkr_mtf = open(self.broker_mtf_directory)
        data = json.load(brkr_mtf)
        broker_list = []
        for item in data:
            broker_list.append([item['ip'],int(item['port']),int(item['timeout'])])
        return broker_list

    def load_topics_to_subscribe(self):
        f = open(self.sub_topics_directory)
        data = json.load(f)
        topic_list = []
        for item in data:
            topic_list.append(item)
        return topic_list
    
    def load_topics_to_publish(self):
        f = open(self.pub_topics_directory)
        data = json.load(f)
        topic_list = []
        for item in data:
            topic_list.append(item)
    
    def Load_clients(self):
        client_mtf = open(self.client_mtf_directory)
        data = json.load(client_mtf)
        c_list = []
        for item in data:
            c_list.append(clients_manager.Client(name=item['name'],
                                                           id=item['id'],
                                                           type=item['type'],
                                                           directory=item['directory'],
                                                           broker=item['broker']))
        return c_list
    
    
    def message_parse(self,msg):
        msg_parts = msg.split("::")
        header_part = msg_parts[0].split('|')
        usr_msg = msg_parts[1]

        if(header_part[2] == "publish_executables"):
            if(header_part[1] == "client_introduction"):
                executables = usr_msg.split(',')
                self.manage_executables(header_part[0],executables_list=executables)
                self.manage_client(header_part[0],executables)
            else:
                self.log("CONTROLLER::CORE:: Publish Executable is sent through unmatching topic. Should be sent through client_introduction.")

        elif(header_part[2] == "echo_name"):            
            self.log("CONTROLLER::CORE:: Echoing " + header_part[0])
        
        elif(header_part[2] == "echo_msg"):            
            self.log("CONTROLLER::CORE:: Echoing " + usr_msg)
        
        else:
            self.log("CONTROLLER::CORE:: No definted Controller function in the message")

    def manage_client(self,client_id,executables):
        for client in self.client_list:
            if (client.id == client_id):
                self.log("CONTROLLER::CORE:: Client already registered.")
                return 0
                
        self.client_list.append(clients_manager.Client(client_id,client_id,"","",self.selected_broker[0],executables=executables))
        
    def manage_executables(self,user_id, executables_list):
        for exec_item in executables_list:
            exists = 0
            for reg_exec_item in self.registered_executables:
                if (reg_exec_item == exec_item):
                    exists = 1
                    break
            if(exists == 0):
                self.registered_executables.append(exec_item)
    

    def MQTT_msg_craft(self,topic,func_name,msg,is_split = False, payload_id = '-1', batch_index = -1):
            payload = "MQTT_FC" + "|" + topic + "|" + func_name + "|" + T.asctime() + "|" + str(is_split) + "|" + payload_id + "|" + str(batch_index) + "|" + "::" + str(msg)
            return payload
        


    def Log(self,msg):
        print(msg)

#____________________________________ THE ONLY BASELINE HARDCODED EXECUTABLES THAT ARE KNOWN TO THE CONTROLLER BY DEFAULT::
    def Pub_Clients_Introduce(self):
        self.client_list.clear()
        msg = self.MQTT_msg_craft("controller_executable","publish_executables","asking to publish your executables")
        print(msg)

        self.client.publish("controller_executable",payload = msg)
    
    def Pub_Clients_Echo_Name(self):
        msg = self.MQTT_msg_craft("controller_executable","echo_name","asking to echo your name")
        print(msg)
        
        self.client.publish("controller_executable",payload = msg) 
##    NOTE: THERE SHOULD BE NO OTHER PRE-KNOWN EXECUTABLES THAT ARE CODED. ONLY INTRODUCED TO THE CONTROLLER DURING RUN TIME.
#___________________________________________________________________________________________________________________________

    def Command_Parser(self, commandline):
        rawText = commandline + ';'
        command_parts = rawText.split(' ')

        if(command_parts[0] == "run"):
            command = command_parts[1].split(';')[0] 
            if(command in self.registered_executables):
                msg = ""
                for i in range(len(command_parts)-1):
                    msg = msg + command_parts[i+1] + " "
                pl = self.MQTT_msg_craft("controller_executable",command,msg)
                self.client.publish("controller_executable",payload = pl) 
                #self.log("CONTROLLER:: executable " + command + " published.")
                self.log("CONTROLLER:: PUBLISHED " + commandline)
            else:
                self.log("CONTROLLER:: executable " + command + " not defined.")

        elif(command_parts[0] == "runp"):
            command = command_parts[1].split(';')[0] 
            self.log("CONTROLLER:: procedure " + command + " initiated.")
            self.my_procedures.parse_procedure_command(command,rawText,self.Command_Parser)
           
        else:
            self.log("CONTROLLER::INPUT>> " + commandline)


from sdflmq import DFLMQ_Coordinator
import time
import sys
import argparse

def main():    
    

    parser = argparse.ArgumentParser(description="SDFLMQ Coordinator quick lunch")

    parser.add_argument('--coordinator_id', type=str, required=True, help='Coordinator id')
    parser.add_argument('--broker_ip', type=str, default="localhost", help='IP address of the broker to connect to')
    parser.add_argument('--broker_port', type=int, default=1883, help='Port number of the broker to connect to')

    args = parser.parse_args()
    
    print(args.coordinator_id)
    print(args.broker_ip)
    print(args.broker_port)
    coordinator_client = DFLMQ_Coordinator( myID            = args.coordinator_id,
                                            broker_ip       = args.broker_ip,
                                            broker_port     = args.broker_port,
                                            loop_forever    = True,
                                            plot_stats      = False)


  
#____________________________________________GUI START_____________________________________________

if __name__ == "__main__":
    main()
	



from .Base.executable_class import PubSub_Base_Executable
from .Base.topics import SDFLMQ_Topics
from .Modules.Coordinator_Modules import components as components
from .Modules.Coordinator_Modules.session_manager import Session_Manager
from .Modules.Coordinator_Modules.clustering_engine import Clustering_Engine
from .Modules.Coordinator_Modules.load_balancer import Load_Balancer
import datetime
import json
import ast
import matplotlib.pyplot as plt
import threading
import random
import time

class DFLMQ_Coordinator(PubSub_Base_Executable) :

    def __init__(self , 
                myID : str , 
                broker_ip : str , 
                broker_port : int , 
                loop_forever : bool,
                plot_stats : bool) -> None : 
        
        topics = SDFLMQ_Topics()
        self.CoTClT = topics.CoTClT # publish 
        self.CiTCoT = topics.ClTCoT # subscribe

        self.session_manager = Session_Manager()
        self.clustering_engine = Clustering_Engine()
        self.load_balancer = Load_Balancer()
        
        self.executables.append('create_fl_session')
        self.executables.append('join_fl_session')
        self.executables.append('leave_fl_session')
        self.executables.append('delete_fl_session')
        self.executables.append('confirm_role')
        self.executables.append('client_received_global')
        self.executables.append('submit_model_stat')

        # self.executables.append('')
        
        super().__init__(
                    myID , 
                    broker_ip , 
                    broker_port ,
                    loop_forever)
            
    def on_connect(self, client, userdata, flags, rc):
        super().on_connect(client, userdata, flags, rc)
        self.client.subscribe(self.CiTCoT,qos=2) 
        print("subscribed to client_to_coordinator topic.")
    
    def __submit_model_stat(self,session_id,client_id,round,acc,loss):
        session = self.session_manager.get_session(session_id)
        # r = session.current_round_index
        r = round
        if(client_id in session.rounds[r]['participants']):
            session.rounds[r]['acc'] = acc
            session.rounds[r]['loss'] = loss
            self.session_manager.save_session_to_file(session_id,self.root_directory)
        else:
            print("client " + str(client_id) + " who wants to submit model stats is not in the list of participants in round " + str(r) + " of sessoin " + str(session_id))

    def __broadcast_roles(self,session_id): #TODO: set retain flag true for this one for other joining clients.
        role_dic = json.dumps(self.session_manager.get_session(session_id).role_dictionary)
        # print(role_dic)
        self.publish(session_id, "set_session_roles"," -s_id " + str(session_id) + " -role_dic " + str(role_dic))
        
    def order_client_resources(self,model_name, dataset_name) : 
        self.publish(self.CoTClT , "echo_resources" , " -model_name " + model_name + " -dataset_name " + dataset_name)
        
    
    def request_client_stats(self,session_id):
        self.publish(topic=session_id,func_name="report_client_stats",msg="")

    def __client_received_global(self,session_id,client_id):
        # print("client " + str(client_id) + " received global for session id: " + str(session_id))
        for c in self.session_manager.get_session(session_id).client_list:
            if(client_id == c.client_id): #If client id matching with root aggregator and if session id matching
                self.session_manager.get_session(session_id).add_participant(client_id)#Increase round step
                if(len(self.session_manager.get_session(session_id).get_participants()) == self.session_manager.get_session(session_id).get_current_round()['num_registered_clients']): #Fix this
                    print("Flagging round as complete.")
                    session_completed_round = self.session_manager.get_session(session_id).get_current_round()
                    self.session_manager.get_session(session_id).complete_round()
                    prcessing_delay = session_completed_round['processing_time']
                    self.session_manager.update_session(session_id)
                    # print(self.session_manager.get_session(session_id).session_status)
                    self.publish(topic=session_id,func_name="round_ack",msg=" -session_id " + str(session_id) + " -ack round_complete")#Send ack to clients "round_complete"
                    if(self.session_manager.get_session(session_id).session_status == components._SESSION_TERMINATED):  #On get_session it is checked if round counter equal to max round of session.
                        print("Session terminated")
                        session = self.session_manager.get_session(session_id)
                        session.session_termination_time = datetime.datetime.now()
                        session.total_processing_time = session.session_termination_time - session.session_creation_time
                        self.session_manager.save_session_to_file(session_id,self.root_directory)
                        self.publish(topic=session_id,func_name="session_ack",msg= " -session_id " + str(session_id) + " -ack_type terminate_s")        #If yes, terminate session, and send ack to clients "session_terminated"
                    elif(self.session_manager.get_session(session_id).session_status == components._SESSION_ACTIVE):#If not, and session is still active then:
                        print("Updating session with new round and updating roles.")
                        
                        self.session_manager.get_session(session_id).new_round()#Set new round
                        self.__update_roles(session_id,prcessing_delay)
                break
            
    # def __round_complete(self, session_id, client_id):
    #     if(client_id == self.session_manager.get_session(session_id).get_root_node().client.client_id): #If client id matching with root aggregator and if session id matching
    #         self.session_manager.get_session(session_id).complete_round()#Increase round step
    #         self.publish(topic=session_id,func_name="round_ack",msg=" -session_id " + str(session_id) + " -ack round_complete")#Send ack to clients "round_complete"
                
    #     if(self.session_manager.get_session(session_id).session_status == components._SESSION_TERMINATED):  #On get_session it is checked if round counter equal to max round of session.
    #         self.publish(topic=session_id,func_name="session_ack",msg= " -session_id " + str(session_id) + " -ack_type terminate_s")        #If yes, terminate session, and send ack to clients "session_terminated"
    #     elif(self.session_manager.get_session(session_id).session_status == components._SESSION_ACTIVE):#If not, and session is still active then:
    #         pass
    #         self.session_manager.get_session(session_id).new_round()#Set new round
    #         self.__update_roles(session_id)#Update roles
            
    def __check_session_status(self,session_id):
        if(self.session_manager.get_session(session_id).session_status == components._SESSION_ACTIVE):#if session ready, clusterize session, and send ack to clients "session ready"
            if(self.session_manager.get_session(session_id).current_round_index == 0):
                self.__clusterize_session(session_id)
            # else:
            #     self.__update_roles(session_id)

            self.publish(topic=session_id,func_name="session_ack",msg=" -session_id " + str(session_id) + " -ack_type active_s")
        elif(self.session_manager.get_session(session_id).session_status == components._SESSION_TIMEOUT):#if min cap not met, and session time is over, terminate session, and send ack to clients
              self.publish(topic=session_id,func_name="session_ack",msg=" -session_id " + str(session_id) + " -ack_type terminate_s")

    def __clusterize_session(self,session_id):
        session = self.session_manager.get_session(session_id)
        self.clustering_engine.create_2layer_topology(session,0.3)
        # self.clustering_engine.create_central_aggregation_topology(session)
        # print("sessions role dictionary: " + str(session.role_dictionary))
        role_dic = json.dumps(session.role_dictionary)
        clusters = self.clustering_engine.form_clusters(session)
        session.set_clusters(clusters)
        # roles_vector = self.load_balancer.initiate_round_robin_no_shuffle(session)
        # roles_vector = self.load_balancer.random_initialize_roles(session)
        roles_vector = self.load_balancer.pso_initialize_roles(session,5)
        # print(roles_vector)
        session.set_roles(roles_vector)
        # print("len nodes: " + str(len(session.nodes)))
        for node in session.nodes:
            # print(node.role)
            if(node.status == components._NODE_PENDING):
                time.sleep(1)
                self.publish(topic=self.CoTClT + node.client.client_id,func_name="set_role",msg= " -s_id " + str(session_id) + " -role " + str(node.role)+ " -role_dic " + str(role_dic))

    def __update_roles(self,session_id,round_processing_delay):
        # print("updating roles")
        session = self.session_manager.get_session(session_id)
        #1) Returns a new role_vector based on the optimizer's suggestion
        #2) Updates the roles according to the new_role_Vector. This only looks into the nodes, and does not need to travers into clusters.
        print("round processing_delay = " + str(round_processing_delay))

        # self.load_balancer.round_robin_no_shuffle(session)
        # self.load_balancer.randomly_update_roles(session)       
        self.load_balancer.pso_optimize_roles(session,round_processing_delay)

        #Inform Clients in nodes with NODE_PENDING status
        no_pending_roles = True
        for node in session.nodes:
            # print(node.role)
            if(node.status == components._NODE_PENDING):
                # time.sleep(1)
                no_pending_roles = False
                print("Notifying clients with new roles\n")
                self.publish(topic=self.CoTClT + node.client.client_id,func_name="reset_role",msg= " -s_id " + str(session_id) + " -role " + str(node.role))
        if(no_pending_roles):
            [ack2,num_active_nodes] = self.session_manager.All_Nodes_Ready(session_id)
            if(ack2):#Check all roles, if all are ready, then send ack to clients "round_ready"
                self.session_manager.get_session(session_id).get_current_round()['num_registered_clients'] = num_active_nodes
                self.publish(topic=session_id,func_name="round_ack",msg=" -session_id " + str(session_id) + " -ack round_ready") 
                # print("published round ready\n")
    
    def __confirm_role(self,session_id,client_id,role):
        # print("confriming role\n")
        ack = self.session_manager.get_session(session_id).confirm_role(role,client_id) #Set new role for the client as ready 
        if(ack == 0):
            [ack2,num_active_nodes] = self.session_manager.All_Nodes_Ready(session_id)
            if(ack2):#Check all roles, if all are ready, then send ack to clients "round_ready"
                self.session_manager.get_session(session_id).get_current_round()['num_registered_clients'] = num_active_nodes
                self.publish(topic=session_id,func_name="round_ack",msg=" -session_id " + str(session_id) + " -ack round_ready") 
                # print("published round ready\n")
       
    def __new_fl_session_request(self,
                                    session_id,
                                    session_time,
                                    session_capacity_min,
                                    session_capacity_max, 
                                    waiting_time, 
                                    model_name,
                                    fl_rounds,
                                    client_id,
                                    client_role,
                                    model_spec,
                                    memcap,
                                    model_size,
                                    pspeed):
        print("Received request for new session\n")
        ack = self.session_manager.create_new_session(session_id,
                                                session_time,
                                                session_capacity_min,
                                                session_capacity_max, 
                                                waiting_time, 
                                                model_name,
                                                model_spec,
                                                fl_rounds)
        
        if(ack == 0):
            self.publish(topic=self.CoTClT + client_id,func_name="session_ack",msg=" -session_id " + str(session_id) + " -ack_type new_s")

        ack2 = self.session_manager.join_session(session_id,
                                                client_id,
                                                client_role,
                                                model_name,
                                                model_spec,
                                                fl_rounds,
                                                memcap,
                                                model_size,
                                                pspeed)
        if(ack2 == 0):
            self.publish(topic=self.CoTClT + client_id,func_name="session_ack",msg=" -session_id " + str(session_id) + " -ack_type join_s")
        
    def __join_fl_session_request(self,
                                    session_id,
                                    client_id,
                                    client_role,
                                    model_name,
                                    model_spec,
                                    fl_rounds,
                                    memcap,
                                    model_size,
                                    pspeed):
        
        ack = self.session_manager.join_session(session_id,
                                            client_id,
                                            client_role,
                                            model_name,
                                            model_spec,
                                            fl_rounds,
                                            memcap,
                                            model_size,
                                            pspeed)
        print("Received request for joining session " + str(session_id) + "\n")
        if(ack == 0):
            self.__check_session_status(session_id)
            self.publish(topic=self.CoTClT + client_id,func_name="session_ack",msg=" -session_id " + str(session_id) + " -ack_type join_s")
    
    def execute_on_msg(self, header_parts, body):
        super().execute_on_msg(header_parts, body) 
        # header_parts = self._get_header_body(msg)
        if header_parts[2] == 'create_fl_session' : 
            client_id = body.split(' -c_id ')[1].split(' -s_id ')[0]
            session_id  = body.split(' -s_id ')[1]  .split(' -s_time ')[0]
            session_time  = body.split(' -s_time ')[1]  .split(' -s_c_min ')[0]
            session_capacity_min  = body.split(' -s_c_min ')[1]    .split(' -s_c_max ')[0]
            session_capacity_max = body.split(' -s_c_max ')[1]   .split(' -waiting_time ')[0]
            waiting_time     = body.split(' -waiting_time ')[1]   .split(' -fl_rounds ')[0]
            fl_rounds     = body.split(' -fl_rounds ')[1]   .split(' -model_name ')[0]
            model_name     = body.split(' -model_name ')[1]   .split(' -model_spec ')[0]
            model_spec     = body.split(' -model_spec ')[1]   .split(' -memcap ')[0]
            memcap     = body.split(' -memcap ')[1]   .split(' -mdatasize ')[0]
            model_size     = body.split(' -mdatasize ')[1]   .split(' -client_role ')[0]
            preferred_role     = body.split(' -client_role ')[1].split(' -pspeed ')[0]
            processing_speed     = body.split(' -pspeed ')[1] .split(';')[0]
            # batch_size     = body.split(' - ')[1]   .split(' - ')[0]
            # num_rounds      = body.split(' - ')[1]    .split(';')[0]
            self.__new_fl_session_request(session_id = session_id,
                                          client_id = client_id,
                                          session_time = session_time,
                                          session_capacity_min = session_capacity_min,
                                          session_capacity_max = session_capacity_max,
                                          waiting_time = waiting_time,
                                          model_name = model_name,
                                          fl_rounds = fl_rounds,
                                          model_spec = model_spec,
                                          memcap = memcap,
                                          model_size = model_size,
                                          client_role = preferred_role,
                                          pspeed = processing_speed)

        if header_parts[2] == 'join_fl_session' :  
            client_id = body.split(' -c_id ')[1]  .split(' -s_id ')[0]
            session_id  = body.split(' -s_id ')[1]  .split(' -fl_rounds ')[0]
            fl_rounds     = body.split(' -fl_rounds ')[1]   .split(' -model_name ')[0]
            model_name     = body.split(' -model_name ')[1]   .split(' -model_spec ')[0]
            model_spec     = body.split(' -model_spec ')[1]   .split(' -memcap ')[0]
            memcap     = body.split(' -memcap ')[1]   .split(' -mdatasize ')[0]
            model_size     = body.split(' -mdatasize ')[1]   .split(' -client_role ')[0]
            preferred_role     = body.split(' -client_role ')[1].split(' -pspeed ')[0]
            processing_speed     = body.split(' -pspeed ')[1] .split(';')[0]
            self.__join_fl_session_request(session_id = session_id,
                                           client_id = client_id,
                                           model_name = model_name,
                                           fl_rounds = fl_rounds,
                                           model_spec = model_spec,
                                           memcap = memcap,
                                           model_size = model_size,
                                           client_role = preferred_role,
                                           pspeed = processing_speed)
        
        if header_parts[2] == 'leave_fl_session' : 
            client_id = body.split(' -c_id ')[1]  .split(' -s_id ')[0]
            session_id  = body.split(' -s_id ')[1]  .split(';')[0]
            print("Leave session has not been implemented yet.")
        
        if header_parts[2] == 'delete_fl_session' : 
            client_id = body.split(' -c_id ')[1]  .split(' -s_id ')[0]
            session_id  = body.split(' -s_id ')[1]  .split(';')[0]
            print("Delete session has not been implemented yet.")
            
        if header_parts[2] == 'confirm_role' : 
            session_id = body.split(' -s_id ')[1]  .split(' -c_id ')[0]
            client_id = body.split(' -c_id ')[1]  .split(' -role ')[0]
            role     = body.split(' -role ')[1].split(';')[0]
            self.__confirm_role(session_id=session_id,
                                client_id=client_id,
                                role=role)
            
        if header_parts[2] == 'client_received_global' : 
            session_id = body.split(' -s_id ')[1]  .split(' -c_id ')[0]
            client_id  = body.split(' -c_id ')[1]  .split(';')[0]
            self.__client_received_global(  session_id=session_id,
                                            client_id=client_id)
            
        if header_parts[2] == 'submit_model_stat' : 
            session_id = body.split(' -s_id ')[1]  .split(' -c_id ')[0]
            client_id  = body.split(' -c_id ')[1]  .split(' -round ')[0]
            round = int(body.split(' -round ')[1]  .split(' -acc ')[0])
            acc = float(body.split(' -acc ')[1]  .split(' -loss ')[0])
            loss  = float(body.split(' -loss ')[1]  .split(';')[0])
            
            self.__submit_model_stat(session_id,client_id,round,acc,loss)
            
        # if header_parts[2] == 'round_complete' : 
        #     session_id = body.split(' -s_id ')[1]  .split(' -c_id ')[0]
        #     client_id  = body.split(' -c_id ')[1]  .split(';')[0]
        #     self.__round_complete(session_id=session_id,
        #                           client_id=client_id)
        
        # print("parsed the message")
